/*

  add subject line to word list and check?
  try with val or test in word list? (what does this mean?)
  change hidden layers?  
  check for bad emails--examine errors

  to start:  keep a simple list of words
	must filter out anything between <>s (assume must begin and end on
		same line?)

  what about URLs?

  presence of certain words (or phrases?  particular words in close proximity)

  matrix of word proximity?  what about multiple occurrences?

  presence of ALL CAPS words or phone numbers or URLS?

  contents of Subject: line, presence of spaces in Subject, numbers in subject

  From: address--presence of numbers?

  presence or absence of certain headers?

  proximity of words to other words

  reject lines with no spaces?  (e.g. base64 encoded)

*/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include "emailnn.h"
#include <stdlib.h>

//mystrcmp(char **s1, char **s2);
int mystrcmp(const void *, const void *);
RemoveHTMLTags(char *buf, FILE *outf, char *wordlist[], int wordlistsize, Email *em);

EmailList *newEmailList()
{
	EmailList *el=(EmailList *) malloc(sizeof(EmailList));

	el->emails=(Email *) malloc(sizeof(Email) * 16);
	el->numemails=0;
	el->numalloced=16;
	return el;
}

void initEmail(Email *em, int wordlistsize)
{
	int i;

	em->numwords=0;
	em->nummatches=0;
	em->wordmatches = (int *) malloc(sizeof(int) * wordlistsize);

	em->filename=(char *) NULL;

	for(i=0; i<wordlistsize; i++) {
		em->wordmatches[i]=0;
	}
}

// return next free email in EmailList
Email *getfreeEmailfromEmailList(EmailList *el, int wordlistsize)
{
	Email *em;
	int ns, nm;

	if(el->numemails==el->numalloced) {
		ns=el->numalloced*2;
		//printf("alloc'd up to %d\n", ns);
		el->numalloced=ns;
		el->emails = (Email *) realloc(el->emails, sizeof(Email)* ns);
	}

	nm=el->numemails;
	el->numemails++;
	em=&(el->emails[nm]);

	initEmail(em, wordlistsize);

	return em;
}


#define BUFSIZE	256

// fname is filename that contains a list of filenames
// go through this list, load each email, extract the
// English words, and print them to outf
PrintWordsInEmailsInFileList(char *fname, FILE *outf)
{
	FILE *f;
	char *buf;

	if(!fname) {
		printf("PrintWordsInEmailsInFileList:  null file\n"); fflush(stdout);
		return -1;
	}

	f=fopen(fname, "r");
	if(!f) {
		printf("failed to open %s\n", fname); fflush(stdout);
		return -1;
	} else {
		printf("PrintWordsInEmailsInFileList:  opened %s\n", fname);
	}

	while(fscanf(f, "%as", &buf) == 1) {
		//printf("PrintWordsInEmailsInFileList:  looking at email '%s'\n", buf); fflush(stdout);
		printf("."); fflush(stdout);
 		EvalEmail(buf, outf, NULL, 0, NULL);
		//printf("-----------------------------------\n");
		free(buf);
	}
	printf("\n");

	fclose(f);
}

EmailList *EvalEmailList(char *fname, char *wordlist[], int wordlistsize)
{
	FILE *f;
	char *buf;
	Email *em=(Email *) NULL;
	EmailList *el=(EmailList *) NULL;

	if(!fname) {
		printf("EvalEmailList:  null file\n");
		return (EmailList *) NULL;
	}

	f=fopen(fname, "r");
	if(!f) {
		printf("failed to open %s\n", fname); fflush(stdout);
		return (EmailList *) NULL;
	} else {
		//printf("opened %s\n", fname);
	}

	el=newEmailList();

	while(fscanf(f, "%as", &buf) != EOF) {
		//printf("EvalEmailList:  looking at email '%s':  ", buf); fflush(stdout);
		em=getfreeEmailfromEmailList(el, wordlistsize);
		em->filename=buf;
		if(strstr(buf, "spam.")!=NULL) {
			em->target=TARGET_HIGH;
		} else if(strstr(buf, "msg.")!=NULL) {
			em->target=TARGET_LOW;
		} else {
			printf("filename '%s' is not classifiable as target\n", em->filename);
			exit(0);
		}
 		EvalEmail(buf, NULL, wordlist, wordlistsize, em);
		//printf("matched %d of %d\n", em->nummatches, em->numwords);
		//free(buf);
	}

	fclose(f);
	return el;
}

EvalEmail(char *fname, FILE *outf, char *wordlist[], int wordlistsize, Email *em)
{
	FILE *f;
	char buf[BUFSIZE+1];
	int inbody=0, nummatchwords=0;

	f=fopen(fname, "r");
	if(!f) {
		printf("failed to open %s\n", fname);
		return -1;
	} else {
		//printf("opened %s\n", fname);
	}

	while(!inbody && fgets(buf, BUFSIZE, f)) {
		//printf("read line of %s", buf);
		if(strlen(buf)==1) {
			inbody=1;
		}
		// new:  look at words in Subject: line too
		if(strncmp(buf, "Subject:", 8) == 0) {
			//printf("found subject line!\n");
			nummatchwords+=RemoveHTMLTags(buf, outf, wordlist, wordlistsize, em);
		}
	}

	if(inbody) {
		while(fgets(buf, BUFSIZE, f)) {
			//printf("body line:  %s", buf); fflush(stdout);
			// replace all text between <> pairs with spaces
			// so that we can parse out individual words
			nummatchwords+=RemoveHTMLTags(buf, outf, wordlist, wordlistsize, em);
		}
	}

	fclose(f);
	return nummatchwords;
}

RemoveHTMLTags(char *buf, FILE *outf, char *wordlist[], int wordlistsize, Email *em)
{
	int found=1, nummatchwords=0;
	char *rb, *rs, *rc;
	char *delim=" \t\r\n*!.?%$#@^():;,\\/[]{}=+-_'&\"";
	static int carryover=0;

	while(found) {
		if(carryover) {		// an HTML tag opened on previous line
			// delete up to >
			rs=buf;
			rb=strchr(buf, (int) '>');	// find first occurrence of HTML tag end	
			if(rc) {
				for(rc=rs; rc<=rb; rc++) { *rc=' '; }
				carryover=0;
			} else {
				// this entire line must be part of HTML tag
				break;	// go to next line, keeping carryover set
			}
		}
		//rb=strpbrk(buf, ">");	// find first occurrence of HTML tag end	
		rb=strchr(buf, (int) '>');	// find first occurrence of HTML tag end	
		if(rb) {
			//printf("found right bracket\n");
			*rb='\0';	// temporarily truncate string
			rs=strrchr(buf, (char) '<');	// find first occurrence of HTML tag start 
			if(rs) {
				// fill <tag> with spaces
				for(rc=rs; rc<=rb; rc++) { *rc=' '; }
			} else {
				// this still happens because some lines just begin with >
				//printf("mismatched HTML tag\n");
				//exit(0);
				// if mismatched, just replace the > with ' ' and try again
				*rb=' ';
			}	
		} else {
			found=0;
		}
	}

	// now see if there is a extra < at end indicating the start of an HTML tag that
	// carries over to next line
	rs=strrchr(buf, (char) '<');	// find first occurrence of HTML tag start 

	// fill to end of line with spaces, and set carryover so next line will
	// have chars deleted to >
	if(rs) {
		carryover=1;
		//printf("setting carryover\n");
		for(rc=rs; *rc!='\0'; rc++) { *rc=' '; }
	}

	// now that the HTML is removed, let's find the words:
	//printf("body after: %s", buf); fflush(stdout);

	rb=strtok(buf, delim);
	while(rb) {
		int l=strlen(rb), i, badchar;
		if((l > 3) && (l < 15) && strcmp(rb, "nbsp")) {
			//printf("'%s' ", rb);
			badchar=0;
			// check for non english chars
			for(i=0; i<l; i++) {
				if(isalpha(rb[i])) {
					rb[i]=tolower(rb[i]);
				} else {
					badchar=1;
					break;
				}
			}
			if(!badchar) {
				//printf("%s\n", rb);
				if(outf) {
					fprintf(outf, "%s\n", rb);
				}
				// we can set numwords to the *total* num of words in email or the number of *matched* words
				//if(em) { em->numwords++; }
				if(wordlist && em) {
					char *m;
					//printf("about to bsearch\n"); fflush(stdout);
					m=bsearch((void *) &rb, (void *) wordlist, wordlistsize, sizeof(char *), mystrcmp);
					//printf("after bsearch\n"); fflush(stdout);
					if(m) {
						int i;

						// printf("match on word %s\n", rb);
						nummatchwords++;
						// if search says word is present, find it:
						// (this is kind of hacky, but bsearch only returns a pointer)
						for(i=0; i<wordlistsize; i++) {
							if(strcmp(rb, wordlist[i]) == 0) {
								//printf("match on index %d\n", i);
								em->wordmatches[i]++;
								em->nummatches++;
								// we can set numwords to the *total* num of words in email or the number of *matched* words
								em->numwords++;
								break;
							}
						}
					}
				}
			}
		}
		rb=strtok(NULL, delim);
	}
	return nummatchwords;
}
