#include <stdio.h>
#include "backprop.h"
#include "emailnn.h"

/*
	make sure no schwab or motion list emails in spam list

	what about x10 in "real"?  it really looks like spam

	need to filter spam and msg lists better

	add Subject: To: and From: to wordlists (?) or make special


*/

// what if we can't find this many valid words in the input emails?
#define NNINPUTSIZE	2000

void LoadWordList(FILE *inf, char *wordlist[]);

main(int argc, char **argv)
{
	int an=1, i, j;
	char *trainsetfn=NULL, *validatesetfn=NULL, *testsetfn=NULL;
	char *errfn;
	EmailList *trainset, *validateset, *testset;
	//char *netfile="network";
	int seed=20020505;
	BPNN *net;
	int epoch, epochs=100;
	double out_err, hid_err, sumerr;
	// these are the same for now, but later when we look at proximity to
	// other words they might not be the same
	int inputnnsize=NNINPUTSIZE;
	int wordlistsize=NNINPUTSIZE;
	char *wordlist[NNINPUTSIZE];
	FILE *wf=(FILE *) NULL, *errfp=(FILE *) NULL;

	while(an < argc) {
		//printf("arg '%s'\n", argv[an]);
		if(strcmp(argv[an], "-train")==0) {
			an++;
			if(an < argc) {
				printf("train set filename is '%s'\n", argv[an]);
				trainsetfn=argv[an];
			}
		} else if(strcmp(argv[an], "-validate")==0) {
			an++;
			if(an < argc) {
				printf("validation set filename is '%s'\n", argv[an]);
				validatesetfn=argv[an];
			}
		} else if(strcmp(argv[an], "-test")==0) {
			an++;
			if(an < argc) {
				printf("test set filename is '%s'\n", argv[an]);
				testsetfn=argv[an];
			}
		} else if(strcmp(argv[an], "-e")==0) {
			an++;
			if(an < argc) {
				epochs=atoi(argv[an]);
				printf("will train for '%d'\n", epochs);
			}
		} else if(strcmp(argv[an], "-errfile")==0) {
			an++;
			if(an < argc) {
				printf("error file '%s'\n", argv[an]);
				errfn=argv[an];
				if(!(errfp=fopen(errfn, "w"))) {
					printf("failed to open error output file\n");
				}
			}
		}
		an++;
	}

	

	//wordlist=CreateGlobalWordList();	// instead of:

	// create a global list of words used in the bodies of all 
	// emails (?--use only trainset?)
	// sort the list by frequency and keep the most commonly used
	// `wordlistsize` words
	// if wordlist.hifreq alread exists, skip creation and load from there
	// to save time
	if(!(wf=fopen("wordlist.hifreq", "r"))) {
		char cmd[1024];
		FILE *outf=fopen("wordlist", "w");

		if(!outf) {
			printf("can't open wordlist for writing\n"); fflush(stdout);
		}
		printf("calling PrintWordsInEmails...\n"); fflush(stdout);
		PrintWordsInEmailsInFileList(trainsetfn, outf);
		//PrintWordsInEmailsInFileList(validateset, outf);	// use this set?
		//PrintWordsInEmailsInFileList(testset, outf);		// use this set?
		fclose(outf);
		sprintf(cmd, "rm -f wordlist.hifreq; cat wordlist | sort | uniq -c | sort -nr | head -%d | awk '{ print $2 }' > wordlist.hifreq", wordlistsize);
		printf("executing command '%s'\n", cmd);
		system(cmd);
		if(!(wf=fopen("wordlist.hifreq", "r"))) {
			printf("main:  couldn't open wordlist.hifreq\n");
			exit(0);
		}
	} else {
		printf("main:  file wordlist.hifreq already exists, using that\n");
	}

	// retrieve the words from file wordlist.hifreq and
	// put in sorted wordlist array:  (still assumes array size known)
	LoadWordList(wf, wordlist);
	fclose(wf);

	// load data about the emails into our local data structures.
	// this needs to be done after the global word list is defined.
	printf("calling EvalEmailList with '%s'\n", trainsetfn); fflush(stdout);
	trainset=EvalEmailList(trainsetfn, wordlist, wordlistsize);

	printf("calling EvalEmailList with '%s'\n", validatesetfn); fflush(stdout);
	validateset=EvalEmailList(validatesetfn, wordlist, wordlistsize);

	printf("calling EvalEmailList with '%s'\n", testsetfn); fflush(stdout);
	testset=EvalEmailList(testsetfn, wordlist, wordlistsize);

	//TestEmailListForWordFrequency(trainsetfn, wordlist, wordlistsize);
	//exit(0);

	bpnn_initialize(seed);

	// future:  could read in a saved network
	// for new, create a new one each time

	net = bpnn_create(inputnnsize, 4, 1);

	for (epoch = 1; epoch <= epochs; epoch++) {
		int showerr;
    	printf("%d ", epoch);  fflush(stdout);

		sumerr = 0.0;
    	for (i = 0; i < trainset->numemails; i++) {
			Email *em=&trainset->emails[i];
			int nw=em->numwords;

      		/** Set up input units on net with input i **/
      		//load_input_with_image(trainlist->list[i], net);
			// for now, input_n must be same size as wordlist
			for(j=0; j<net->input_n; j++) {
				if(em->numwords==0) {
					net->input_units[j]=0.0;
				} else {
					// dividing by nm gives a sort of "weight" adjustment based on
					net->input_units[j]=(double)em->wordmatches[j]/(double) nw;
					// try with no weight:  (seems to perform abysmally)
					// net->input_units[j]=(double)em->wordmatches[j];
				}
			}

      		/** Set up target vector for image i **/
      		//load_target(trainlist->list[i], net);
			net->target[1]=em->target;

      		/** Run backprop, learning rate 0.3, momentum 0.3 **/
			bpnn_train(net, 0.3, 0.3, &out_err, &hid_err);

      		sumerr += (out_err + hid_err);
    	}

		// see how we are doing
		if(epoch==epochs) showerr=1;
		else showerr=0;

		TestPerf(trainset, net, showerr, errfp);
		TestPerf(validateset, net, showerr, errfp);
		TestPerf(testset, net, showerr, errfp);

    	//printf("%g ", sumerr);

    	printf("\n");  fflush(stdout);

	}
	printf("\n"); fflush(stdout);


	fclose(errfp);
}

int mystrcmp(char **s1, char **s2)
{
	//printf("comparing %s and %s\n", *s1, *s2);
	return strcmp(*s1, *s2);
}

void LoadWordList(FILE *inf, char *wordlist[])
{
	int i=0, fs;

	printf("loading words\n");
	while((fs=fscanf(inf, "%as", &wordlist[i])) == 1) {
		//printf("loaded word %s\n", wordlist[i]);
		i++;
	}
	//printf("fs was %d\n", fs);

	//qsort(base, nmemb, size, compare);
	printf("LoadWordList:  quicksorting %d words\n", i);
	qsort(wordlist, i, sizeof(char *), mystrcmp);

#if 0
	{
	int j;
	printf("sorted words:\n");
	for(j=0; j<i; j++) {
		printf("%s\n", wordlist[j]);
	}	
	}
#endif
}

#if 0
TestEmailListForWordFrequency(char *trainsetfn, char *wordlist, int wordlistsize)
{
	char *efn;
	FILE *inf=fopen(trainsetfn, "r");
	int nummatchwords;

	if(!inf) {
		printf("TestEmailListForWordFrequency:  failed to open %s\n", trainsetfn);
		return 0;
	}

	while(fscanf(inf, "%as", &efn)==1) {
		printf("Matching words in '%s'\n", efn);
		nummatchwords=EvalEmail(efn, NULL, wordlist, wordlistsize);
		free(efn);
		printf("%d\n", nummatchwords);
	}

	fclose(inf);
}
#endif

void TestPerf(EmailList *set, BPNN *net, int showerr, FILE *errfp)
{
	int correct=0, i, j;
	double err=0.0, delta;

	if(!set) return;

	for (i = 0; i < set->numemails; i++) {
		Email *em=&set->emails[i];
		int nw=em->numwords;

		/** Set up input units on net with input i **/
		//load_input_with_image(trainlist->list[i], net);
		// for now, input_n must be same size as wordlist
		for(j=0; j<net->input_n; j++) {
			if(em->numwords==0) {
				net->input_units[j]=0.0;
			} else {
				net->input_units[j]=(double)em->wordmatches[j]/(double) nw;
			}
		}

		bpnn_feedforward(net);

		net->target[1]=em->target;
		
		delta=(net->target[1] - net->output_units[1]);
		err+=0.5*delta*delta;

		if (net->target[1] > 0.5 && net->output_units[1] > 0.5) {
			correct++;
		} else if (net->target[1] < 0.5 && net->output_units[1] < 0.5) {
			correct++;
		} else if(showerr) {
			//printf("error on %s\n", em->filename);
			fprintf(errfp, "%s\n", em->filename);
		}

	}
	printf("\t%2.4g\t%2.4g ", ((double) correct / (double) set->numemails) * 100.0, err);
}
